#ifndef _RECTANGLE_
#define _RECTANGLE_

class CRectangle
{
private:
	double m_dLeft;
	double m_dRight;
	double m_dTop;
	double m_dBottom;

public:
	inline CRectangle();
	inline CRectangle(
		const double dLeft, 
		const double dRight,
		const double dTop, 
		const double dBottom
	);
	inline CRectangle(const C2DVector &cCenter, C2DVector &cSize);

	inline double GetTopY() const;
	inline double GetBottomY() const;
	inline double GetRightX() const;
	inline double GetLeftX() const;

	inline const C2DVector GetTopLeftCorner() const;
	inline const C2DVector GetTopRightCorner() const;
	inline const C2DVector GetBottomLeftCorner() const;
	inline const C2DVector GetBottomRightCorner() const;

	inline const C2DLine GetLeftEdge() const;
	inline const C2DLine GetRightEdge() const;
	inline const C2DLine GetTopEdge() const;
	inline const C2DLine GetBottomEdge() const;
	
	inline double GetWidth() const;
	inline double GetHeight() const;
	inline const C2DVector GetCenter() const;

	inline bool IsIn(const C2DVector &cPoint) const;
	inline const C2DVector Adjust(C2DVector cPoint) const;
	inline double DistanceToEdge(const C2DVector &cPoint) const;

	inline const CRectangle ExpandLeft(const double dValue) const;
	inline const CRectangle ExpandRight(const double dValue) const;
	inline const CRectangle ExpandTop(const double dValue) const;
	inline const CRectangle ExpandBottom(const double dValue) const;
	inline const CRectangle ShrinkLeft(const double dValue) const;
	inline const CRectangle ShrinkRight(const double dValue) const;
	inline const CRectangle ShrinkTop(const double dValue) const;
	inline const CRectangle ShrinkBottom(const double dValue) const;

	inline bool operator ==(const CRectangle &cdValue) const;

};

CRectangle::CRectangle()
:	m_dLeft(0.0),
	m_dRight(0.0),
	m_dTop(0.0),
	m_dBottom(0.0)
{
}

CRectangle::CRectangle(
	const double dLeft, 
	const double dRight,
	const double dTop, 
	const double dBottom
)
:	m_dLeft(dLeft),
	m_dRight(dRight),
	m_dTop(dTop),
	m_dBottom(dBottom)
{
}

CRectangle::CRectangle(const C2DVector &cCenter, C2DVector &cSize)
{
	cSize *= 0.5;

	m_dLeft = cCenter.GetX() - cSize.GetX();
	m_dRight = cCenter.GetX() + cSize.GetX();
	m_dTop = cCenter.GetY() - cSize.GetY();
	m_dBottom = cCenter.GetY() + cSize.GetY();
};

double CRectangle::GetTopY() const
{
	return m_dTop;
}

double CRectangle::GetBottomY() const
{
	return m_dBottom;
}

double CRectangle::GetRightX() const
{ 
	return m_dRight;
}

double CRectangle::GetLeftX() const
{ 
	return m_dLeft;
}

const C2DVector CRectangle::GetTopLeftCorner() const
{ 
	return C2DVector(m_dLeft, m_dTop); 
}

const C2DVector CRectangle::GetTopRightCorner() const
{
	return C2DVector(m_dRight, m_dTop);
}

const C2DVector CRectangle::GetBottomLeftCorner() const
{
	return C2DVector(m_dLeft, m_dBottom);
}

const C2DVector CRectangle::GetBottomRightCorner() const
{ 
	return C2DVector(m_dRight, m_dBottom);
}

const C2DLine CRectangle::GetLeftEdge() const 
{
	return C2DLine(GetTopLeftCorner(), GetBottomLeftCorner());
}

const C2DLine CRectangle::GetRightEdge() const
{
	return C2DLine(GetTopRightCorner(), GetBottomRightCorner());
}

const C2DLine CRectangle::GetTopEdge() const
{
	return C2DLine(GetTopLeftCorner(), GetTopRightCorner());
}

const C2DLine CRectangle::GetBottomEdge() const 
{
	return C2DLine(GetBottomLeftCorner(), GetBottomRightCorner());
}

double CRectangle::GetWidth() const
{ 
	return m_dRight - m_dLeft;
}

double CRectangle::GetHeight() const
{
	return m_dBottom - m_dTop;
}

const C2DVector CRectangle::GetCenter() const
{
	return C2DVector(m_dLeft + 0.5 * GetWidth(), m_dTop + 0.5 * GetHeight());
}

bool CRectangle::IsIn(const C2DVector& cPoint) const
{
	return 
		(cPoint.GetX() >= m_dLeft) && 
		(cPoint.GetX() <= m_dRight) &&
		(cPoint.GetY() >= m_dTop) && 
		(cPoint.GetY() <= m_dBottom);
}

double CRectangle::DistanceToEdge(const C2DVector &cPoint) const
{
	if(!IsIn(cPoint))
		// distance outside is a negative number
		return -(Adjust(cPoint) - cPoint).GetDistance();

	return 
		min((cPoint.GetX() - m_dLeft), min((m_dRight - cPoint.GetX()),
			min((cPoint.GetY() - m_dTop), (m_dBottom - cPoint.GetY()))));	  
}

const C2DVector CRectangle::Adjust(C2DVector cPoint) const
{
	if(cPoint.GetX() < m_dLeft)
		cPoint.SetY(m_dLeft);
	else if(cPoint.GetX() > m_dRight)
		cPoint.SetY(m_dRight);

	if(cPoint.GetY() < m_dTop)
		cPoint.SetY(m_dTop);
	else if(cPoint.GetY() > m_dBottom)
		cPoint.SetY(m_dBottom);

	return cPoint;
};

const CRectangle CRectangle::ExpandLeft(const double dValue) const
{
	return CRectangle(m_dLeft - dValue, m_dRight, m_dTop, m_dBottom);
}

const CRectangle CRectangle::ExpandRight(const double dValue) const
{
	return CRectangle(m_dLeft, m_dRight + dValue, m_dTop, m_dBottom);
}

const CRectangle CRectangle::ExpandTop(const double dValue) const
{
	return CRectangle(m_dLeft, m_dRight, m_dTop - dValue, m_dBottom);
}

const CRectangle CRectangle::ExpandBottom(const double dValue) const
{
	return CRectangle(m_dLeft, m_dRight, m_dTop, m_dBottom + dValue);
}

const CRectangle CRectangle::ShrinkLeft(const double dValue) const
{
	return ExpandLeft(-dValue);
}

const CRectangle CRectangle::ShrinkRight(const double dValue) const
{
	return ExpandRight(-dValue);
}

const CRectangle CRectangle::ShrinkTop(const double dValue) const
{ 
	return ExpandTop(-dValue);
}
const CRectangle CRectangle::ShrinkBottom(const double dValue) const
{
	return ExpandBottom(-dValue);
}

bool CRectangle::operator ==(const CRectangle &cdValue) const
{
	return 
		(m_dLeft == cdValue.m_dLeft) && 
		(m_dRight == cdValue.m_dRight) &&
		(m_dTop == cdValue.m_dTop) && 
		(m_dBottom == cdValue.m_dBottom); 
}

#endif  // _RECTANGLE_
